"""Base Processor for messages consumed from a channel."""
import logging
from abc import ABC
from abc import abstractmethod
from typing import Type

import pika
from pydantic import BaseModel
from pydantic import Field
from pydantic import ValidationError

from talus.models.message import ConsumeMessageBase

logger = logging.getLogger(__name__)


DEFAULT_DEAD_LETTER_EXCEPTIONS = (ValidationError,)


class MessageProcessorBase(BaseModel, ABC):
    message_cls: Type[ConsumeMessageBase]
    dead_letter_exceptions: Type[Exception] | tuple[Type[Exception]] = Field(
        default=DEFAULT_DEAD_LETTER_EXCEPTIONS
    )

    @abstractmethod
    def process_message(self, message: ConsumeMessageBase):
        pass  # pragma: no cover

    @classmethod
    def dlq_message(
        cls,
        channel: pika.adapters.blocking_connection.BlockingChannel,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        exception: Exception | None = None,
    ) -> None:
        logger.warning(f"Dead lettering message:{method=}, {properties=}, {exception=}")
        channel.basic_nack(delivery_tag=method.delivery_tag, requeue=False)

    @classmethod
    def acknowledge_message(
        cls, channel: pika.adapters.blocking_connection.BlockingChannel, message: ConsumeMessageBase
    ) -> None:
        logger.debug(
            f"Acknowledging message: method={message.method}, properties={message.properties}, {message=}"
        )
        channel.basic_ack(delivery_tag=message.delivery_tag)

    def __call__(
        self,
        channel: pika.adapters.blocking_connection.BlockingChannel,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ) -> "MessageProcessorBase":
        try:
            message = self.message_cls(method=method, properties=properties, body=body)
            self.process_message(message)
            self.acknowledge_message(channel=channel, message=message)
        except self.dead_letter_exceptions as e:
            self.dlq_message(channel=channel, method=method, properties=properties, exception=e)
        return self


class DLQMessageProcessor(MessageProcessorBase):
    message_cls: Type[ConsumeMessageBase] = ConsumeMessageBase
    dead_letter_exceptions: Type[Exception] | tuple[Type[Exception]] = Field(
        default_factory=lambda: DEFAULT_DEAD_LETTER_EXCEPTIONS + (ValueError,)
    )

    def process_message(self, message: ConsumeMessageBase):
        raise ValueError(f"DLQMessageProcessor auto fails processing: {message=}")


class MessageProcessorSelectorBase(BaseModel, ABC):
    message_processors: dict[str, MessageProcessorBase]

    @abstractmethod
    def get_message_processor(
        self,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ) -> MessageProcessorBase:
        pass  # pragma: no cover

    def __call__(
        self,
        channel: pika.adapters.blocking_connection.BlockingChannel,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ) -> MessageProcessorBase:
        message_processor = self.get_message_processor(
            method=method, properties=properties, body=body
        )
        return message_processor(channel=channel, method=method, properties=properties, body=body)


class MessageProcessorRoutingKeySelector(MessageProcessorSelectorBase):
    def get_message_processor(
        self,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ) -> MessageProcessorBase:
        routing_key = method.routing_key

        message_processor = self.message_processors.get(routing_key)
        if message_processor is None:
            logger.error(f"No message processor found for routing key: {routing_key}")
            message_processor = DLQMessageProcessor()
        return message_processor



"""Queue model."""
from pydantic import BaseModel


class Queue(BaseModel):
    """Queue model."""

    name: str
    passive: bool = False
    durable: bool = True
    auto_delete: bool = False
    exclusive: bool = False
    arguments: dict | None = None  # TODO https://www.rabbitmq.com/docs/quorum-queues#declaring

    def __str__(self):
        return self.name



"""
Module encapsulating message structure to facilitate use with the Consumer and Producer wrappers
"""
import uuid
from typing import Type

import pika.spec
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field


class MessageBodySchemaBase(BaseModel):
    """
    Base class for message schemas which can be used to validate message bodies.
    """

    model_config = ConfigDict(extra="allow")

    conversation_id: str = Field(default_factory=lambda: uuid.uuid4().hex, alias="conversationId")


class MessageBase:
    """
    Base class for messages establishing a common interface for use by DurableConnections
    """

    message_body_schema_cls: Type[MessageBodySchemaBase] = MessageBodySchemaBase

    def __init__(self, routing_key: str, body: bytes | dict | str | MessageBodySchemaBase):
        self.routing_key = routing_key
        if isinstance(body, dict | BaseModel):
            self.body: MessageBodySchemaBase = self.message_body_schema_cls.model_validate(body)
        else:  # str | bytes
            self.body: MessageBodySchemaBase = self.message_body_schema_cls.model_validate_json(
                body
            )


class ConsumeMessageBase(MessageBase):
    """
    Base class for messages consumed from a DurableConsumer MessageProcessor.
    TODO example of how to use this class
    """

    def __init__(
        self,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes | dict | str | MessageBodySchemaBase,
    ):
        super().__init__(method.routing_key, body)
        self.method = method
        self.properties = properties

    @property
    def delivery_tag(self):
        return self.method.delivery_tag

    @property
    def headers(self) -> dict | None:
        return self.properties.headers


class PublishMessageBase(MessageBase):
    """
    Base class for messages published to with a DurableProducer
    """

    routing_key: str = "default.m"
    headers: dict[str, str] | None = (None,)

    def __init__(
        self,
        body: bytes | dict | str | MessageBodySchemaBase,
    ):
        super().__init__(self.routing_key, body)

    @property
    def properties(self) -> pika.spec.BasicProperties:
        return pika.BasicProperties(
            content_type="text/plain",
            priority=0,
            delivery_mode=pika.DeliveryMode.Persistent,
            content_encoding="UTF-8",
            headers=self.headers,
        )



"""Tests for the message processor models."""
from typing import Type
from uuid import uuid4

import pika
import pytest

from talus.models.message import ConsumeMessageBase
from talus.models.processor import DLQMessageProcessor
from talus.models.processor import MessageProcessorBase
from talus.models.processor import MessageProcessorRoutingKeySelector


@pytest.fixture()
def successful_message_processor(consume_message_cls) -> MessageProcessorBase:
    """
    A message processor that does not raise an exception.
    """

    class SuccessfulMessageProcessor(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls

        def process_message(self, message: ConsumeMessageBase):
            self._message_was_processed = True

    yield SuccessfulMessageProcessor()


@pytest.fixture()
def consumed_message(
    publish_message_cls,
    producer,
    consumer,
    body,
) -> tuple[pika.spec.Basic.Deliver, pika.spec.BasicProperties, bytes]:
    outbound_message = publish_message_cls(body=body)
    producer.publish(message=outbound_message)
    for m, p, b in consumer.consume_generator(auto_ack=True):
        yield m, p, b  # only the first one
        break


def test_message_processor_success(consumer, consumed_message, successful_message_processor):
    """
    :given: A message processor that does not raise an exception.
    :when: The message processor is called.
    :then: The message processor should process and acknowledge the message.
    """
    # given
    method, properties, body = consumed_message
    # when
    successful_message_processor(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert successful_message_processor._message_was_processed
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue


@pytest.fixture()
def erroring_message_processor(consume_message_cls) -> MessageProcessorBase:
    """
    A message processor that raises a non-dead-letter exception.
    """

    class ErroringMessageProcessor(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls
        dead_letter_exceptions: Type[Exception] = ValueError

        def process_message(self, message: ConsumeMessageBase):
            raise RuntimeError("A bad message processor raised an exception that is not DLQ'd.")

    yield ErroringMessageProcessor()


def test_message_processor_failure(erroring_message_processor, consumed_message, consumer):
    """
    :given: A message processor that raises an exception.
    :when: The message processor is called.
    :then: Expected exception is raised.
    """
    # given
    method, properties, body = consumed_message
    # when/then
    with pytest.raises(RuntimeError):
        erroring_message_processor(
            channel=consumer.channel, method=method, properties=properties, body=body
        )


@pytest.fixture()
def dlq_message_processor(consume_message_cls) -> DLQMessageProcessor:
    """
    A message processor that raises a non-dead-letter exception.
    """

    class AutoDLQMessageProcessor(DLQMessageProcessor):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls
        dead_letter_exceptions: Type[Exception] = RuntimeError

        def process_message(self, message: ConsumeMessageBase):
            self._raised_run_time_error = True
            raise RuntimeError("A bad message processor raised an exception that is DLQ'd.")

    yield AutoDLQMessageProcessor()


def test_dlq_message_processor(dlq_message_processor, consumed_message, consumer):
    """
    :given: A message processor that raises an exception identified as for DLQ.
    :when: The message processor is called.
    :then: The message processor should DLQ the message.
    """
    # given
    method, properties, body = consumed_message
    # when
    dlq_message_processor(channel=consumer.channel, method=method, properties=properties, body=body)
    # then
    assert dlq_message_processor._raised_run_time_error
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue


@pytest.fixture()
def message_processor_routing_key_selector(
    routing_key, successful_message_processor
) -> MessageProcessorRoutingKeySelector:
    """
    A message processor selector that selects a message processor based on the routing key.
    """
    return MessageProcessorRoutingKeySelector(
        message_processors={
            routing_key: successful_message_processor,
        }
    )


def test_message_processor_routing_key_selector_found(
    message_processor_routing_key_selector, consumed_message, consumer, successful_message_processor
):
    """
    :given: A message processor selector that selects a message processor based on the routing key.
    :when: The message processor selector is called with a message of that routing key.
    :then: The expected message processor should be run and returned.
    """
    # given
    method, properties, body = consumed_message
    # when
    message_processor = message_processor_routing_key_selector(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue
    assert isinstance(message_processor, type(successful_message_processor))


def test_message_processor_routing_key_selector_not_found(
    message_processor_routing_key_selector, consumer, consumed_message
):
    """
    :given: A message processor selector that selects a message processor based on the routing key.
    :when: The message processor selector is called with a message an unknown routing key.
    :then: The DLQ message processor should be returned.
    """
    # given
    method, properties, body = consumed_message
    unknown_routing_key = uuid4().hex
    method.routing_key = unknown_routing_key
    # when/then
    message_processor = message_processor_routing_key_selector(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue
    assert isinstance(message_processor, DLQMessageProcessor)
